# Copyright 2018 Yegor Bitensky

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from distutils.core import setup, Extension


setup(
    name='CTHPoker',
    version='1.0.1',
    description="Texas Hold'em Poker combination finder tool",
    url='https://github.com/YegorDB/CTHPoker',
    author='Yegor Bitensky',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Topic :: Games/Entertainment",
        "Topic :: Education",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: C",
    ],
    ext_modules=[Extension(name='CTHPoker', sources=['finder.c'])]
)
